/*
 * $Id$
 *
 * Copyright (c) 2014, Juniper Networks, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see
 * <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
 */

#include <string.h>
#include <stdio.h>
#include <stddef.h>

#include "atavendor_db.h"

#define UNIGEN_KCI_SERIES              "UGB94BPH"
#define UNIGEN_KC_SERIES               "UGB94ARF"
#define VITRIUM_STORFLY_SERIES         "StorFly - VSFA18PI"
#define VITRIUM_EM_SERIES              "VSFA18PI"
#define STEC_MACH_SERIES               "STEC MACH-8 SSD"
#define STEC_M8SB1A_SERIES             "M8SB1A"
#define SOLIGEN_13S_SERIES             "SGC13S"
#define SOLIGEN_13T_SERIES             "SGC13T"

static const struct ata_vendor_db *  getdrivedb(char *model);
static u_int64_t getrawvalue(unsigned char *raw);
static int genericchecksmart(struct ata_smart_values data,
			      struct smart_ssd_thresholds thres);
static int unigenchecksmart(struct ata_smart_values smartdata,
			      struct smart_ssd_thresholds thres);

const struct ata_vendor_attributes unigen_attr[] = {
    {   1 , "Raw Read Error Rate     "},
    {   9 , "Power On Hours Count    "},
    {  12 , "Power Cycle Count       "},
    { 184 , "Initial Bad Block Count "},
    { 195 , "Program Failure Block Ct"},
    { 196 , "Erase Failure Block Ct  "},
    { 197 , "Read Failure Block Ct   "},
    { 198 , "Total Read Sectors Ct   "},
    { 199 , "Total Write Sectors Ct  "},
    { 200 , "Total Read Commands Ct  "},
    { 201 , "Total Write Commands Ct "},
    { 202 , "Total Err Bits Flash Ct "},
    { 203 , "Rd Sect correc Bit err  "},
    { 204 , "Bad Block Full Flag     "},
    { 205 , "Max PE Count Spec       "},
    { 206 , "Erase Count Min         "},
    { 207 , "Erase Count Max         "},
    { 208 , "Erase Count Average     "},
    { 209 , "Remaining Life (%)      "},
    { 210 , "BBM Error Log           "},
    { 000 , "END OF VENDOR ATTRIBUTE "}
};

const struct ata_vendor_attributes storfly_attr[] = {
    {   1 , "Raw Read Error Rate     "},
    {   5 , "Reallocated Sector Ct.  "},
    {   9 , "Power On Hours Count    "},
    {  12 , "Power Cycle Count       "},
    { 160 , "Uncorrectable Sector Ct "},
    { 161 , "No of Valid Spare Block "},
    { 163 , "No of Init Inv Block    "},
    { 164 , "Total Erase Count       "},
    { 165 , "Maximum Erase Count     "},
    { 166 , "Minimum Erase Count     "},
    { 167 , "Average Erase Count     "},
    { 192 , "Power Off Retract Count "},
    { 194 , "Device Temperature      "},
    { 195 , "Hardware ECC Recovered  "},
    { 196 , "Reallocation Event Ct.  "},
    { 198 , "Offline Uncorrectable   "},
    { 199 , "UDMA CRC Error Count    "},
    { 241 , "Total LBAs Written      "},
    { 242 , "Total LBAs Read         "},
    { 000 , "END OF VENDOR ATTRIBUTE "}
};

const struct ata_vendor_attributes vit_em_attr[] = {
    {   1 , "Raw Read Error Rate     "},
    {   5 , "Reallocated Sector Ct.  "},
    {   9 , "Power On Hours Count    "},
    {  12 , "Power Cycle Count       "},
    { 160 , "Uncorrectable Sector Ct "},
    { 161 , "No of Valid Spare Block "},
    { 163 , "No of Init Inv Block    "},
    { 164 , "Total Erase Count       "},
    { 165 , "Maximum Erase Count     "},
    { 166 , "Minimum Erase Count     "},
    { 167 , "Average Erase Count     "},
    { 192 , "Power Off Retract Count "},
    { 194 , "Device Temperature      "},
    { 195 , "Hardware ECC Recovered  "},
    { 196 , "Reallocation Event Ct.  "},
    { 198 , "Offline Uncorrectable   "},
    { 199 , "UDMA CRC Error Count    "},
    { 241 , "Total LBAs Written      "},
    { 242 , "Total LBAs Read         "},
    { 243 , "Remaining Life (%)      "},
    { 000 , "END OF VENDOR ATTRIBUTE "}
};

const struct ata_vendor_attributes stec_attr[] = {
    {   1 , "Raw Read Error Rate     "},
    {   5 , "Reallocated Sector Ct.  "},
    {   9 , "Power On Hours Count    "},
    {  12 , "Power Cycle Count       "},
    {  13 , "Soft Read Error Rate    "},
    { 100 , "Erase/Program Cycle Ct  "},
    { 170 , "Reserved Block Count    "},
    { 171 , "Program Fail Count      "},
    { 172 , "Erase Fail Count        "},
    { 173 , "Wear Leveling Count     "},
    { 174 , "Unexpected Power Loss   "},
    { 183 , "Runtime Bad Block       "},
    { 184 , "End-to-End Err Detect   "},
    { 187 , "Hard Errors             "},
    { 195 , "ECC On-the-Fly          "},
    { 199 , "UDMA CRC Error Count    "},
    { 000 , "END OF VENDOR ATTRIBUTE "}
};

const struct ata_vendor_attributes soligen_attr[] = {
    {   1 , "Raw Read Error Rate     "},
    {   5 , "Reallocated Sector Ct.  "},
    {   9 , "Power On Hours Count    "},
    {  12 , "Power Cycle Count       "},
    {  13 , "Soft Read Error Rate    "},
    { 100 , "Gigabytes Erased        "},
    { 170 , "Reserved Block Count    "},
    { 171 , "Program Fail Count      "},
    { 172 , "Erase Fail Count        "},
    { 174 , "Unexpected Power Loss   "},
    { 177 , "Wear Range Delta        "},
    { 181 , "Program Fail Count      "},
    { 182 , "Erase Fail Count        "},
    { 184 , "IOEDC Errors            "},
    { 187 , "Reported Uncorrect Err  "},
    { 194 , "Device Temperature      "},
    { 195 , "ECC On-the-Fly          "},
    { 196 , "Reallocation Event Ct.  "},
    { 198 , "Uncorrectable Sector Ct "},
    { 199 , "SATA R-Errors Error Ct  "},
    { 201 , "Uncorrect Soft Read Err "},
    { 204 , "Soft ECC Correction Rate"},
    { 230 , "Life Curve Status       "},
    { 231 , "SSD Life Left           "},
    { 232 , "Available Reserved Space"},
    { 233 , "Data Written to Flash   "},
    { 234 , "Data Written by Host    "},
    { 235 , "Supercap Monitoring     "},
    { 241 , "Lifetime Wr from Host   "},
    { 242 , "Lifetime Rd from Host   "},
    { 000 , "END OF VENDOR ATTRIBUTE "}
};

const struct ata_vendor_attributes default_attr[] = {
    {   1 , "Raw Read Error Rate     "},
    {   2 , "Throughput Performance  "},
    {   3 , "Spin Up Time            "},
    {   4 , "Start Stop Count        "},
    {   5 , "Reallocated Sector Ct.  "},
    {   7 , "Seek Error Rate         "},
    {   8 , "Seek Time Preformance   "},
    {   9 , "Power On Hours Count    "},
    {  10 , "Spin Retry Count        "},
    {  12 , "Power Cycle Count       "},
    { 191 , "Gsense Error Rate       "},
    { 192 , "Power Off Retract Count "},
    { 193 , "Load/Unload Cycle Count "},
    { 194 , "Device Temperature      "},
    { 196 , "Reallocation Event Ct.  "},
    { 197 , "Current Pending Sec. Ct "},
    { 198 , "Offline Uncorrectable   "},
    { 199 , "UDMA CRC Error Count    "},
    { 200 , "Write Error Rate        "},
    { 000 , "END OF VENDOR ATTRIBUTE "}
};

/* Vendor attributes database and vendor series table */

const struct ata_vendor_db vendors_db[] = {
    {UNIGEN_KCI_SERIES,       unigen_attr,   unigenchecksmart  },
    {UNIGEN_KC_SERIES,        unigen_attr,   unigenchecksmart  },
    {VITRIUM_STORFLY_SERIES,  storfly_attr,  genericchecksmart },
    {VITRIUM_EM_SERIES,       vit_em_attr,   genericchecksmart },
    {STEC_M8SB1A_SERIES,      stec_attr,     genericchecksmart },
    {STEC_MACH_SERIES ,       stec_attr,     genericchecksmart },
    {SOLIGEN_13S_SERIES,      soligen_attr,  genericchecksmart },
    {SOLIGEN_13T_SERIES,      soligen_attr,  genericchecksmart }
};

/* size of vendor database table */
#define SIZE_OF_VENDORS_DB  ((sizeof(vendors_db))/(sizeof(vendors_db[0])))


const struct ata_vendor_attributes *vendor_attributes_db = NULL;
static int unigen_ssd = 0;

int
isvendorattribexists(void)
{
    return (vendor_attributes_db ? 1 : 0);
}

/* The function get the model name scans the database and
 * returns the pointer to data for the model if it exists
 * in database, else returns NULL
 */
static const struct ata_vendor_db *
getdrivedb(char *model)
{
    unsigned int i = 0;
    const char *vendor_series = NULL;

    while (i < SIZE_OF_VENDORS_DB) {
	vendor_series = vendors_db[i].vendor_series_name;
	if (strstr(model, vendor_series))
	    break;
	i++;
    }

    if (i >= SIZE_OF_VENDORS_DB) {
	return NULL;
    } else {
	if ((strcmp(vendor_series, UNIGEN_KCI_SERIES) == 0) ||
	    (strcmp(vendor_series, UNIGEN_KC_SERIES) == 0)) {
	    unigen_ssd = 1;
	}
    }

    return (&vendors_db[i]);
}

void
atavendorattributesdbupdate(char *model)
{
    const struct ata_vendor_db *vendor_db = getdrivedb(model);

    if (vendor_db) {
	vendor_attributes_db = vendor_db->vendor_attributes;
    } else {
	vendor_attributes_db = NULL;
    }

}

#define ATTR_SIZE 30
static char vendor_attribute_name[ATTR_SIZE];

char *
atagetvendorattribname(unsigned char id)
{
    int i;
    for (i = 0; ((i < NUMBER_ATA_SMART_ATTRIBUTES) &&
                (vendor_attributes_db[i].id != 0)); i++) {
	if ( vendor_attributes_db[i].id == id) {
	    snprintf(vendor_attribute_name, ATTR_SIZE, "(%3d)%s ", id,
		     vendor_attributes_db[i].vendor_attrib_name);
	    return vendor_attribute_name ;
	}
    }
    if (id >= 201) {
	snprintf(vendor_attribute_name, ATTR_SIZE,
		 "(%3d)Vendor Unique           ", id);
    } else {
	snprintf(vendor_attribute_name, ATTR_SIZE,
		 "(%3d)Unknown Attribute       ", id);
    }
    return vendor_attribute_name ;
}

int
isunigenattributetbl(void)
{
    return unigen_ssd;
}

int
printsmartvalues(void *ptr, int smartdesc, int pos)
{
    struct ata_smart_attribute_unigen *data =
	(struct ata_smart_attribute_unigen *) ptr;

    switch (smartdesc) {

	case STATUS_FLAG :
	    return data->status_flag;
	    break;

	case NORMAL_VALUE :
	    return data->normalized;
	    break;

	case WORST_VALUE :
	    return data->worse_normal;
	    break;

	case RAW_VALUE :
	    return data->raw[pos];
	    break;

	default :
	    return 0;
	    break;
    }
}

/* The function returns the raw value */
static u_int64_t
getrawvalue(unsigned char *raw)
{
    int i;
    u_int64_t rawvalue = 0;

    for (i = 5; i >= 0 ; i--) {
	rawvalue = rawvalue << 8;
	rawvalue |= raw[i];
    }
    return rawvalue;
}

/* the function unigenchecksmart implements unigen specific
 * smart check algorithm.
 * Attribute ID 209 - Remaining Life (%) configured in smart.conf
 * a)	This attribute is normalized to 100 in the Raw Value field.
 * b)	This number decreases.
 * c)	Read Raw Value (in hex)
 * return value :
 * id(209) :  Case 1: If Attribute 195 + Attribute 196 >= 0x8.
 *                    If the Raw Value field decreases to 0x14 (20 decimal).
 *                    This indicates 20% remaining spare blocks.
 *            Case 2: If Attribute 195 + Attribute 196 < 0x8.
 *                    If the Raw Value field decreases to 0x5.
 *                    This is indicates 5% remaining life.
 * 0       : if raw value is with the limits mentioned above.
 */
static int
unigenchecksmart(struct ata_smart_values smartdata,
		 struct smart_ssd_thresholds thres)
{
    unsigned char i;
    u_int64_t val = 0, rawval_195 = 0, rawval_196 = 0;
    struct ata_smart_attribute_unigen *data;
    unsigned int thresl;


    for (i = 0 ; i < NUMBER_ATA_SMART_ATTRIBUTES ; i++) {

	switch (smartdata.vendor_attributes[i].id ) {
	    case 195:
		/* raw value of the attribute 195 */
		data = (struct ata_smart_attribute_unigen*)
		    &(smartdata.vendor_attributes[i]);
		rawval_195 = getrawvalue(data->raw);
		break;
	    case 196:
		/* raw value of the attribute 196 */
		data = (struct ata_smart_attribute_unigen*)
		    &(smartdata.vendor_attributes[i]);
		rawval_196 = getrawvalue(data->raw);
		break;
	    default:
		/* check only for the ids which are configured in smartd.conf */
		if (IS_FLAG_VALID(thres, i)) {
		    data = (struct ata_smart_attribute_unigen*)
			&(smartdata.vendor_attributes[i]);

		    /* whether to check raw or normalised values for ids */
		    if(IS_FLAG_RAW(thres, i)) {
			val = getrawvalue(data->raw);
		    } else {
			val = data->normalized;
		    }

		    /* the two threshold limits for the id are stored in
		     * threshold as low 16 bits for 1st threshold and
		     * high 16bits for the 2nd thershold
		     */
		    if (rawval_195 + rawval_196 >= 0x08) {
			thresl = (thres.thres_entries[i].threshold) & 0xFFFF;
		    } else {
			thresl = ((thres.thres_entries[i].threshold) >> 16)
			    & 0xFFFF;
		    }

		    /* check against high or low threshold limits */
		    if(IS_FLAG_LOW(thres, i)) {
			if (val <= thresl) {
			    return thres.thres_entries[i].id;
			}
		    } else {
			if (val >= thresl) {
			    return thres.thres_entries[i].id;
			}
		    }
		}
		break;
	}
    }
    return 0;
}

/* The function genericchecksmart is for the ssd's which
 * doesn't have vendor specific algorithm and for future
 * ssd's. the complete information of the threshold check
 * have to added in the smartd.conf for the function to
 * work properly
 * return value :
 *  0   :if attribute values within the thresholds limits
 *  id  :the id of the attribute which in not in the threshold limit
 */
static int
genericchecksmart(struct ata_smart_values data,
		  struct smart_ssd_thresholds thres)
{
    unsigned char i;
    u_int64_t val;

    for (i = 0 ; i < NUMBER_ATA_SMART_ATTRIBUTES ; i++) {
	/* check only for the ids which are configured in smartd.conf */
	if (IS_FLAG_VALID(thres, i)) {
	    /* whether to check raw or normalised values for ids */
	    if(IS_FLAG_RAW(thres, i)) {
		val = getrawvalue(data.vendor_attributes[i].raw);	
	    } else {
		val = data.vendor_attributes[i].normalized;
	    }
	    /* check against high or low threshold limits */
	    if(IS_FLAG_LOW(thres, i)) {
		if (val <= thres.thres_entries[i].threshold) {
		    return thres.thres_entries[i].id;
		}
	    } else {
		if (val >= thres.thres_entries[i].threshold) {
		    return thres.thres_entries[i].id;
		}
	    }
	}
    }
    return 0;
}

/* The function ssdchecksmart will check the ssd smart values against thershold
 * values from the smartd.conf, by calling vendor specific smartcheck.
 * return value :
 *  0   :if attribute values within the thresholds limits
 *  id  :the id of the attribute which in not in the threshold limit
 */
int
ssdchecksmart(struct ata_smart_values data, struct smart_ssd_thresholds thres)
{
    unsigned char id = 0;
    const struct ata_vendor_db *vendor_db = getdrivedb(thres.ssd_model);

    atavendorattributesdbupdate(thres.ssd_model);
    if (vendor_db && vendor_db->checksmart) {
	id = vendor_db->checksmart(data, thres);
    } else {
	id = genericchecksmart(data, thres);
    }
    return id;
}

